import scipy
import numpy as np
import pandas as pd
import anndata as ad


def read_table(filename: str, sep: str = "\t") -> pd.DataFrame:
    """
    Reads an input file and confirms that
    the file is in an expected format
    """
    frame = pd.read_csv(
        filename,
        header=None,
        names=["barcode", "guide", "n_umi"],
        dtype={"barcode": str, "guide": str, "n_umi": int},
        sep=sep,
    )
    matrix = frame.pivot_table(
        index="barcode", columns="guide", values="n_umi", fill_value=0
    )
    return matrix


def read_anndata(filename: str) -> pd.DataFrame:
    """
    Reads an .h5ad formatted file and
    confirms that the file is in an expected format
    """
    adata = ad.read(filename)
    if isinstance(adata.X, scipy.sparse._csr.csr_matrix):
        mat = adata.X.todense()
    else:
        mat = adata.X
    matrix = pd.DataFrame(
        mat,
        index=adata.obs.index.values,
        columns=adata.var.index.values,
    )
    return matrix



from geomux import read_table, read_anndata, Geomux


def test_anndata_example():
    """
    loads an existing anndata and processes it
    """
    matrix = read_anndata("example/example.h5ad")
    gx = Geomux(
        matrix, cell_names=matrix.index.values, guide_names=matrix.columns.values
    )
    gx.test()
    assignments = gx.assignments()
    assert assignments.shape[0] == matrix.shape[0]


def test_anndata_sparse_csr_example():
    """
    loads an existing anndata and processes it
    """
    matrix = read_anndata("example/example_sparse.h5ad")
    gx = Geomux(
        matrix, cell_names=matrix.index.values, guide_names=matrix.columns.values
    )
    gx.test()
    assignments = gx.assignments()
    assert assignments.shape[0] == matrix.shape[0]


def test_table_example():
    """
    loads an existing table and processes it
    """
    matrix = read_table("example/example.tsv.gz")
    gx = Geomux(
        matrix, cell_names=matrix.index.values, guide_names=matrix.columns.values
    )
    gx.test()
    assignments = gx.assignments()
    assert assignments.shape[0] == matrix.shape[0]



from multiprocessing import Pool
from typing import List, Union, Optional
import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
from scipy.stats import hypergeom
from scipy.special import logit
from adjustpy import adjust


class Geomux:
    def __init__(
        self,
        matrix: Union[np.ndarray, pd.DataFrame],
        cell_names: Optional[Union[List[str], np.ndarray, ArrayLike]] = None,
        guide_names: Optional[Union[List[str], np.ndarray, ArrayLike]] = None,
        min_umi: int = 5,
        min_cells: int = 100,
        n_jobs: int = 4,
        verbose: bool = False,
        method: str = "bh",
    ):
        """
        Parameters
        ----------
        matrix : np.ndarray
            matrix of cell x guide counts
        min_umi : int
            minimum number of UMIs to consider a cell barcode
        min_cells : int
            minimum number of cells to consider a guide
        n_jobs : int
            number of jobs to use for multiprocessing
        method: str
            pvalue adjustment procedure to use.
        """

        # Load the matrix
        if isinstance(matrix, pd.DataFrame):
            matrix = matrix.values
        self.matrix = matrix

        # Load the cell and guide names
        if cell_names is None:
            cell_names = np.arange(matrix.shape[0])
        else:
            assert len(cell_names) == matrix.shape[0]
            cell_names = np.array(cell_names)

        if guide_names is None:
            guide_names = np.arange(matrix.shape[1])
        else:
            assert len(guide_names) == matrix.shape[1]
            guide_names = np.array(guide_names)

        self.cell_names = cell_names
        self.guide_names = guide_names

        # Set the parameters
        self.min_umi = min_umi
        self.min_cells = min_cells
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.method = method
        self._n_total = matrix.shape[0]
        self._m_total = matrix.shape[1]

        self._set_procedure()
        self._filter_matrix()
        self._fit_parameters()

        self._n_cells = self.matrix.shape[0]
        self._n_guides = self.matrix.shape[1]
        self._n_tests = self._n_cells * self._n_guides

        self.is_fit = False
        self.labels = []

    def _set_procedure(self):
        allowed_procedures = ["bonferroni", "bh", "by"]
        if self.method not in allowed_procedures:
            raise ValueError(
                f"Provided method {self.method} not recognized. Choose from {', '.join(allowed_procedures)}"
            )

    def _filter_matrix(self):
        """
        Filters the matrix to only include cells with at least
        `min_umi` UMIs
        """
        cell_sums = self.matrix.sum(axis=1)
        guide_sums = self.matrix.sum(axis=0)
        self.passing_cells = cell_sums >= self.min_umi
        self.passing_guides = guide_sums >= self.min_cells
        self.matrix = self.matrix[self.passing_cells][:, self.passing_guides]
        if self.verbose:
            print("---------------------------------------")
            print(
                "INFO: Average number of UMIs per cell: {:.2f}".format(cell_sums.mean())
            )
            print("INFO: Variance of UMIs per cell: {:.2f}".format(cell_sums.var()))
            print(
                "INFO: Cell UMI counts range from {} to {}".format(
                    cell_sums.min(), cell_sums.max()
                )
            )
            print("---------------------------------------")

            print(
                "INFO: Average number of cells per guide: {:.2f}".format(
                    guide_sums.mean()
                )
            )
            print("INFO: Variance of cells per guide: {:.2f}".format(guide_sums.var()))
            print(
                "INFO: Guide cell counts range from {} to {}".format(
                    guide_sums.min(), guide_sums.max()
                )
            )
            print("---------------------------------------")
            old_cell_size = self._n_total
            new_cell_size = self.matrix.shape[0]
            print(
                "LOG: Removed {} ({:.2f}%) cells with < {} UMIs".format(
                    old_cell_size - new_cell_size,
                    100 * (old_cell_size - new_cell_size) / old_cell_size,
                    self.min_umi,
                )
            )

            old_guide_size = self._m_total
            new_guide_size = self.matrix.shape[1]
            print(
                "LOG: Removed {} ({:.2f}%) guides with < {} Cells".format(
                    old_guide_size - new_guide_size,
                    100 * (old_guide_size - new_guide_size) / old_guide_size,
                    self.min_cells,
                )
            )
        if self.matrix.shape[0] == 0:
            raise ValueError(
                "No cells passed the UMI threshold. Try lowering the min_umi parameter"
            )
        if self.matrix.shape[1] == 0:
            raise ValueError(
                "No guides passed the cell threshold. Try lowering the min_cells parameter"
            )

    def _fit_parameters(self):
        """
        Fits the hypergeometric distribution parameters
        """
        self.population = self.matrix.sum()
        self.successes = self.matrix.sum(axis=0)
        self.draws = self.matrix.sum(axis=1)

    def _hypergeometric_test(self, x: np.ndarray, idx: int):
        """
        Calculates the survival function of a hypergeometric
        distribution for each cell x guide pair.

        Parameters
        ----------
        x : np.ndarray
            vector of guide counts for a single cell
        idx : int
            index of the cell in the matrix
        """
        return hypergeom.sf(x, self.population, self.successes, self.draws[idx])

    def _adjust_pvalues(self, pvalues: np.ndarray):
        """
        Adjust pvalues using the Bonferroni correction
        """
        adj_pvalues = adjust(pvalues, self.method)
        adj_pvalues = np.clip(adj_pvalues, np.min(adj_pvalues[adj_pvalues != 0]), 1)
        return adj_pvalues.reshape(pvalues.shape)

    def _log_odds(self):
        """
        calculates log odds as the ratio between the majority
        and second majority guide in a cell.

        In the case of negative-log pvalues this will calculate
        the log odds ratio between the observed significance
        """
        lor = np.zeros(self._n_cells)
        for i in np.arange(self._n_cells):
            maj_idx, min_idx = np.argsort(self.pv_mat[i])[:2]
            maj_val, min_val = self.pv_mat[i][[min_idx, maj_idx]]
            lor[i] = logit(maj_val) - logit(min_val)
        return lor

    def test(self):
        """
        Performs cell x guide geometric testing
        """
        with Pool(self.n_jobs) as p:
            pv_mat = np.vstack(
                p.starmap(
                    self._hypergeometric_test,
                    zip(self.matrix, np.arange(self._n_cells)),
                )
            )

        pv_mat = np.clip(pv_mat, np.min(pv_mat[pv_mat != 0]), 1)
        self.pv_mat = self._adjust_pvalues(pv_mat)
        self.log_odds = self._log_odds()
        self.is_fit = True
        return self.pv_mat

    def _calc_assignments(self, threshold=0.05) -> List:
        """
        Predict significant assignments for each cell
        """
        if not self.is_fit:
            AttributeError("Please run `.test()` method first")
        guide_indices = np.arange(self._m_total)
        guide_mask = guide_indices[self.passing_guides]
        self.labels = [
            self.guide_names[guide_mask[np.flatnonzero(self.pv_mat[i] < threshold)]]
            for i in np.arange(self._n_cells)
        ]
        self.counts = [
            self.matrix[i][np.flatnonzero(self.pv_mat[i] < threshold)]
            for i in np.arange(self._n_cells)
        ]
        self.pvalues = [
            self.pv_mat[i][np.flatnonzero(self.pv_mat[i] < threshold)]
            for i in np.arange(self._n_cells)
        ]
        return self.labels

    def _calc_moi(self, threshold=0.05):
        """
        Classify each cell between single, double, or null assignments
        """
        if not self.is_fit:
            AttributeError("Please run `.test()` method first")
        self.classification = np.sum(self.pv_mat < threshold, axis=1)
        return self.classification

    def assignments(self, threshold=0.05):
        """
        Returns a dataframe for all assignments with significance
        """
        cell_id_in = np.arange(self._n_total)[self.passing_cells]
        cell_id_out = np.arange(self._n_total)[~self.passing_cells]

        cell_name_in = self.cell_names[self.passing_cells]
        cell_name_out = self.cell_names[~self.passing_cells]

        self._calc_assignments(threshold)

        frame = pd.DataFrame(
            {
                "cell_id": cell_name_in,
                "assignment": self.labels,
                "counts": self.counts,
                "pvalues": self.pvalues,
                "moi": self._calc_moi(threshold),
                "n_umi": self.draws,
                "min_pvalue": self.pv_mat.min(axis=1),
                "max_count": self.matrix.max(axis=1),
                "log_odds": self.log_odds,
                "tested": True,
            },
            index=cell_id_in,
        )
        null = pd.DataFrame(
            {
                "cell_id": cell_name_out,
                "assignment": [
                    np.array([]) for _ in np.arange(np.sum(~self.passing_cells))
                ],
                "counts": [
                    np.array([]) for _ in np.arange(np.sum(~self.passing_cells))
                ],
                "pvalues": [
                    np.array([]) for _ in np.arange(np.sum(~self.passing_cells))
                ],
                "moi": np.nan,
                "n_umi": np.nan,
                "min_pvalue": np.nan,
                "max_count": np.nan,
                "log_odds": np.nan,
                "tested": False,
            },
            index=cell_id_out,
        )
        return pd.concat([frame, null]).sort_index()



import sys
import argparse
from geomux import Geomux, read_table, read_anndata


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i", "--input", type=str, required=True, help="Input table to assign"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        help="output table of barcode assignments (default=stdout)",
    )
    parser.add_argument(
        "-u",
        "--min_umi",
        type=int,
        required=False,
        default=5,
        help="minimum number of UMIs to consider a cell (default=5)",
    )
    parser.add_argument(
        "-c",
        "--min_cells",
        type=int,
        required=False,
        default=100,
        help="minimum number of cells to consider a guide (default=100)",
    )
    parser.add_argument(
        "-t",
        "--threshold",
        type=float,
        required=False,
        default=0.05,
        help="Pvalue threshold to use after pvalue correction (default=0.05)",
    )
    parser.add_argument(
        "-C",
        "--correction",
        type=str,
        required=False,
        default="bh",
        help="Pvalue correction method to use (default=bh)",
    )
    parser.add_argument(
        "-j",
        "--n_jobs",
        type=int,
        required=False,
        default=1,
        help="Number of jobs to use when calculating hypergeometric distributions (default=1)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        required=False,
        help="Suppress progress messages",
    )
    args = parser.parse_args()
    return args


def main_cli():
    args = get_args()

    if args.input.endswith(".h5ad"):
        matrix = read_anndata(args.input)
    else:
        matrix = read_table(args.input)

    if args.correction not in ["bh", "bonferroni", "by"]:
        raise ValueError("Correction method must be one of: bh, bonferroni, by")

    gx = Geomux(
        matrix,
        cell_names=matrix.index.values,
        guide_names=matrix.columns.values,
        min_umi=args.min_umi,
        min_cells=args.min_cells,
        n_jobs=args.n_jobs,
        method=args.correction,
        verbose=not args.quiet,
    )
    gx.test()
    assignments = gx.assignments(threshold=args.threshold)

    if not args.output:
        assignments.to_csv(sys.stdout, sep="\t", index=False)
    else:
        assignments.to_csv(args.output, sep="\t", index=False)



import numpy as np
import pandas as pd
from geomux import Geomux
from muxsim import MuxSim


def sample_barcode(b_size):
    """
    generates a sample barcode
    """
    return "".join(np.random.choice(["A", "C", "T", "G"], size=b_size))


def generate_matrix(n, b_size, g_size):
    frame = pd.DataFrame(
        {
            "barcode": [sample_barcode(b_size) for _ in np.arange(n)],
            "guide": [sample_barcode(g_size) for _ in np.arange(n)],
            "n_umi": [np.random.choice(100) for _ in np.arange(n)],
        }
    )
    matrix = frame.pivot_table(index="barcode", columns="guide", values="n_umi").fillna(
        0
    )
    return matrix


def test_geomux():
    """
    tests basic usage
    """
    n = 100
    b_size = 10
    g_size = 4
    matrix = generate_matrix(n, b_size, g_size)
    gx = Geomux(matrix, min_cells=1)
    gx.test()
    assignments = gx.assignments()
    assert assignments.shape[0] == n


def test_assignments():
    """
    tests assignments with easy data
    """
    num_cells = 1000
    num_guides = 100
    ms = MuxSim(
        num_cells=num_cells,
        num_guides=num_guides,
        n=20,
    )
    gen = ms.sample()
    gx = Geomux(gen)
    gx.test()
    assignments = gx.assignments()
    guide_assignments = assignments.assignment.apply(
        lambda x: max(x) if len(x) > 0 else -1
    ).max()
    assert assignments.shape[0] == num_cells
    assert guide_assignments <= num_guides
    assert assignments.moi.min() >= 0
    assert (assignments.moi == 0).sum() > 0
    assert (assignments.moi == 1).sum() > 0
    assert (assignments.moi == 2).sum() > 0


def test_geomux_min_cells():
    """
    tests min_cells
    """
    num_cells = 1000
    num_guides = 100
    ms = MuxSim(
        num_cells=num_cells,
        num_guides=num_guides,
        n=20,
    )
    gen = ms.sample()
    gen[:, :3] = 0
    gx = Geomux(gen, min_cells=5)
    gx.test()
    assignments = gx.assignments()
    num_guides = np.sum(gen.sum(axis=0) >= 5)
    assert gx._n_guides == num_guides
    for a in assignments.assignment:
        for i in [0, 1, 2]:
            assert i not in a


def test_geomux_all_cells_filtered():
    """
    tests conditions where all cells are filtered
    """
    gen = np.zeros((100, 100))
    try:
        gx = Geomux(gen, min_umi=5)
        assert False
    except ValueError:
        pass


def test_geomux_all_guides_filtered():
    """
    tests conditions where all guides are filtered
    """
    gen = np.ones((100, 100))
    try:
        gx = Geomux(gen, min_cells=101)
        assert False
    except ValueError:
        pass


def test_geomux_correct_assignment():
    """
    tests that the correct assignment is made for certain cases
    """
    num_cells = 1000
    num_guides = 100
    ms = MuxSim(
        num_cells=num_cells,
        num_guides=num_guides,
        n=20,
    )
    gen = ms.sample()
    gen[:, :3] = 0
    gx = Geomux(gen, min_cells=5)
    gx.test()
    assignments = gx.assignments()

    for exp, obs in zip(ms.assignments, assignments.assignment):
        if 3 in exp:
            assert 3 in obs
        elif 4 in exp:
            assert 4 in obs



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="geomux",
    version="0.2.5",
    author="Noam Teysser",
    author_email="Noam.Teyssier@ucsf.edu",
    packages=["geomux"],
    description="a tool to assign an identity to a table of barcode guides",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/noamteyssier/geomux",
    entry_points={"console_scripts": ["geomux = geomux.__main__:main_cli"]},
    install_requires=[
        "numpy",
        "pandas",
        "anndata",
        "scipy",
        "adjustpy",
    ],
)
